from typing import Any

from pydantic import BaseModel, Field, model_validator

from azuma import types
from azuma.parsers import apply_modifiers, normalize_field_map, prepare_condition


class DetectionField(BaseModel):
    list_search: list[types.Query] = Field(default_factory=list)
    map_search: list[types.DetectionMap] = Field(default_factory=list)


def normalize_field_block(name: str, field: Any) -> DetectionField:
    if isinstance(field, dict):
        return DetectionField(map_search=[normalize_field_map(field)])

    if isinstance(field, list):
        if all(isinstance(_x, dict) for _x in field):
            return DetectionField(map_search=[normalize_field_map(_x) for _x in field])

        return DetectionField(
            list_search=[apply_modifiers(str(_x), ["contains"]) for _x in field]
        )

    raise ValueError(f"Failed to parse selection field {name}: {field}")


def normalize_detection(detection: dict[str, Any]) -> dict[str, DetectionField]:
    return {name: normalize_field_block(name, data) for name, data in detection.items()}


class Detection(BaseModel):
    detection: dict[str, DetectionField] = Field(...)
    timeframe: str | None = Field(default=None)
    condition: types.Condition = Field(...)

    @model_validator(mode="before")
    @classmethod
    def transform(cls, values: Any):
        timeframe: str | None = None
        if "timeframe" in values:
            timeframe = values.pop("timeframe")

        condition = prepare_condition(values.pop("condition"))
        detection = normalize_detection(values)
        return {"condition": condition, "detection": detection, "timeframe": timeframe}

    def get_search_fields(self, search_id: str) -> DetectionField | None:
        return self.detection.get(search_id)

    @property
    def all_searches(self) -> dict[str, DetectionField]:
        return self.detection



from typing import Literal

from pydantic import BaseModel, Field


class Related(BaseModel):
    id: str = Field(...)
    type: Literal["derived", "obsoletes", "merged", "renamed", "similar"] = Field(...)



import json
from pathlib import Path
from typing import Any

import yaml
from pydantic import BaseModel

YAML_CONTENT_TYPES = ["text/yaml", "application/x-yaml"]


class YAMLBaseModel(BaseModel):
    """BaseModel with YAML support"""

    @classmethod
    def parse_file(  # type: ignore
        cls,
        path: str | Path,
        *,
        content_type: str | None = "text/yaml",
        encoding: str = "utf8",
        strict: bool | None = None,
        context: dict[str, Any] | None = None,
    ):
        """Parse a file.

        Args:
            path (str | Path): Path to a file.
            content_type (str | None, optional): Content type. Defaults to "text/yaml".
            encoding (str, optional): Encoding. Defaults to "utf8".
            strict (bool | None, optional): Strict or not. Defaults to None.
            context (dict[str, Any] | None, optional): Context. Defaults to None.

        Returns:
            YAMLBaseModel: Parsed instance.
        """
        with open(path, encoding=encoding) as f:
            text = f.read()

            if content_type in YAML_CONTENT_TYPES:
                obj = yaml.safe_load(text)
            else:
                obj = json.loads(text)

        return cls.model_validate(obj, strict=strict, context=context)

    @classmethod
    def parse_raw(  # type: ignore
        cls,
        b: str | bytes,
        *,
        content_type: str | None = "text/yaml",
        strict: bool | None = None,
        context: dict[str, Any] | None = None,
    ):
        """Parse a raw text.

        Args:
            b (str | bytes): String or bytes.
            content_type (str | None, optional): Content type. Defaults to "text/yaml".
            strict (bool | None, optional): Strict or not. Defaults to None.
            context (dict[str, Any] | None, optional): Context. Defaults to None.

        Returns:
            YAMLBaseModel: Parsed instance.
        """
        if content_type in YAML_CONTENT_TYPES:
            obj = yaml.safe_load(b)
        else:
            obj = json.loads(b)

        return cls.model_validate(obj, strict=strict, context=context)



"""
This parser uses lark to transform the condition strings from signatures into callbacks that
invoke the right sequence of searches into the rule and logic operations.
"""

import pathlib
import typing as t

from lark import Lark, Token, Transformer, Tree

from azuma.exceptions import UnsupportedFeature
from azuma.matchers import analyze_x_of, match_search_id

BASE_PATH = pathlib.Path(__file__).parent
GRAMMAR_PATH = BASE_PATH / "../grammar.lark"

grammar = GRAMMAR_PATH.read_text()


class FactoryTransformer(Transformer):
    @staticmethod
    def start(args):
        return args[0]

    @staticmethod
    def search_id(args: list[Token]):
        name = args[0].value

        def match_hits(signature, event):
            return match_search_id(signature, event, name)

        return match_hits

    @staticmethod
    def search_pattern(args: list[Token]):
        return args[0].value

    @staticmethod
    def atom(args: list[t.Callable]):
        if not all(callable(x) for x in args):
            raise ValueError(args)

        return args[0]

    @staticmethod
    def not_rule(args: list[t.Any]):
        negate, value = args

        negate = t.cast(Tree | None, negate)
        value = t.cast(t.Callable, value)

        assert callable(value)

        if negate is None:
            return value

        def _negate(*state):
            return not value(*state)

        return _negate

    @staticmethod
    def and_rule(args: list[t.Callable]):
        if not all(callable(x) for x in args):
            raise ValueError(args)

        if len(args) == 1:
            return args[0]

        def _and_operation(*state):
            for component in args:
                if not component(*state):
                    return False

            return True

        return _and_operation

    @staticmethod
    def or_rule(args: list[t.Any]):
        if not all(callable(x) for x in args):
            raise ValueError(args)

        if len(args) == 1:
            return args[0]

        def _or_operation(*state):
            for component in args:
                if component(*state):
                    return True

            return False

        return _or_operation

    @staticmethod
    def pipe_rule(args: t.Any):
        return args[0]

    @staticmethod
    def x_of(args: t.Any):
        # Load the left side of the X of statement
        count = None
        if args[0].children[0].type == "NUMBER":
            count = int(args[0].children[0].value)

        # Load the right side of the X of statement
        selector: str | None = str(args[2])
        if selector == "them":
            selector = None

        # Create a closure on our
        def _check_of_sections(signature, event):
            return analyze_x_of(signature, event, count, selector)

        return _check_of_sections

    @staticmethod
    def aggregation_expression(args):
        raise UnsupportedFeature("Aggregation expressions not supported.")

    @staticmethod
    def near_aggregation(args):
        raise UnsupportedFeature("Near operation not supported.")


# Create & initialize Lark class instance
factory_parser = Lark(
    grammar, parser="lalr", transformer=FactoryTransformer(), strict=True
)


def prepare_condition(raw_condition: str | list) -> Tree:
    if isinstance(raw_condition, list):
        raw_condition = "(" + ") or (".join(raw_condition) + ")"

    return factory_parser.parse(raw_condition)
